# Copyright (C) 2009-2011 W. Trevor King <wking@drexel.edu>
#
# This file is part of pyafm.
#
# pyafm is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the
# Free Software Foundation, either version 3 of the License, or (at your
# option) any later version.
#
# pyafm is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with pyafm.  If not, see <http://www.gnu.org/licenses/>.

"""Tools for controlling atomic force microscopes.

Provides control of AFM postition using both short-range (piezo) and
long range (stepper) vertical positioning.  There are separate modules
for controlling the piezo (`pypiezo`) and stepper (`stepper`), this
module only contains methods that require the capabilities of both.
"""

from pypiezo.base import convert_bits_to_meters as _convert_bits_to_meters
from pypiezo.base import convert_meters_to_bits as _convert_meters_to_bits
from pypiezo.base import convert_volts_to_bits as _convert_volts_to_bits
from pypiezo.surface import FlatFit as _FlatFit
from pypiezo.surface import SurfaceError as _SurfaceError

from . import LOG as _LOG


class AFM (object):
    """Atomic force microscope positioning.

    Uses a short range `piezo` and a long range `stepper` to position
    an AFM tip relative to the surface.

    Parameters
    ----------
    piezo | pypiezo.afm.AFMpiezo instance
        Fine positioning and deflection measurements.
    stepper | stepper.Stepper instance
        Coarse positioning.
    temperature | temperature.Controller instance or None
        Optional temperature monitoring and control.
    """
    def __init__(self, piezo, stepper, temperature=None, axis_name='z'):
        self.piezo = piezo
        self.stepper = stepper
        self.temperature = temperature
        self.axis_name = axis_name

    def get_temperature(self):
        """Measure the sample temperature.

        Return the sample temperature in Kelvin or `None` if such a
        measurement is not possible.
        """
        if hasattr(self.temperature, 'get_temperature'):
            return self.temperature.get_temperature()

    def move_just_onto_surface(self, depth=-50e-9, setpoint=2,
                               min_slope_ratio=10, far=200):
        """Position the AFM tip close to the surface.

        Uses `.piezo.get_surface_position()` to pinpoint the position
        of the surface.  Adjusts the stepper position as required via
        `.stepper.step_relative()` to get within
        `2*.stepper.step_size` meters of the surface.  Then adjusts
        the piezo to place the cantilever `depth` meters onto the
        surface.  Negative `depth`\s place the tip off the surface

        If `.piezo.get_surface_position()` fails to find the surface,
        backs off `far` half steps (for safety) and steps in (without
        moving the zpiezo) until deflection voltage is greater than
        `setpoint`.
        """
        _LOG.info('moving to %g onto the surface' % depth)

        stepper_tolerance = 2*self.stepper.step_size

        axis = self.piezo.axis_by_name(self.axis_name)

        zero = _convert_volts_to_bits(axis.config['channel'], 0)
        target_def = _convert_volts_to_bits(axis.config['channel'], setpoint)

        _LOG.debug('zero the %s piezo output' % self.axis_name)
        self.piezo.jump(axis_name=self.axis_name, position=zero)

        _LOG.debug("see if we're starting near the surface")
        try:
            pos = self.piezo.get_surface_position(
                axis_name=self.axis_name, max_deflection=target_def,
                min_slope_ratio=min_slope_ratio)
        except _FlatFit, e:
            _LOG.info(e)
            pos = self._stepper_approach_again(
                target_deflection=target_def, min_slope_ratio=min_slope_ratio,
                far=far)
        except _SurfaceError, e:
            _LOG.info(e)
            pos = self._stepper_approach_again(
                target_deflection=target_def, min_slope_ratio=min_slope_ratio,
                far=far)

        pos_m = _convert_bits_to_meters(axis.config, pos)
        _LOG.debug('located surface at stepper %d, piezo %d (%g m)'
                  % (self.stepper.position, pos, pos_m))

        _LOG.debug('fine tune the stepper position')
        while pos_m < -stepper_tolerance:  # step back if we need to
            _LOG.debug('step back')
            self.stepper.step_relative(-1, backlash_safe=True)
            try:
                pos = self.piezo.get_surface_position(
                    axis_name=self.axis_name, max_deflection=target_def,
                    min_slope_ratio=min_slope_ratio)
            except _FlatFit, e:
                _LOG.debug(e)
                continue
            pos_m = _convert_bits_to_meters(axis.config, pos)
            _LOG.debug('located surface at stepper %d, piezo %d (%g m)'
                      % (self.stepper.position, pos, pos_m))
        while pos_m > stepper_tolerance:  # step forward if we need to
            _LOG.debug('step forward')
            self.stepper.step_relative(1)
            try:
                pos = self.piezo.get_surface_position(
                    axis_name=self.axis_name, max_deflection=target_def,
                    min_slope_ratio=min_slope_ratio)
            except _FlatFit, e:
                _LOG.debug(e)
                continue
            pos_m = _convert_bits_to_meters(axis.config, pos)
            _LOG.debug('located surface at stepper %d, piezo %d (%g m)'
                      % (self.stepper.position, pos, pos_m))

        _LOG.debug('adjust the %s piezo to place us just onto the surface'
                  % self.axis_name)
        target_m = pos_m + depth
        target = _convert_meters_to_bits(axis.config, target_m)
        self.piezo.jump(self.axis_name, target)

        _LOG.debug(
            'positioned %g m into the surface at stepper %d, piezo %d (%g m)'
            % (depth, self.stepper.position, target, target_m))


    def _stepper_approach_again(self, target_deflection, min_slope_ratio, far):
        _LOG.info('back off %d half steps and approach until deflection > %g'
                 % (far, target_deflection))
        # back away
        self.stepper.step_relative(-far, backlash_safe=True)
        self.stepper_approach(target_deflection=target_deflection)
        for i in range(2*max(1, self.stepper.backlash)):
            _LOG.debug(
                'additional surface location attempt (stepping backwards)')
            try:
                pos = self.piezo.get_surface_position(
                    axis_name=self.axis_name, max_deflection=target_deflection,
                    min_slope_position=min_slope_postion)
                return pos
            except _SurfaceError, e:
                _LOG.info(e)
            self.stepper.single_step(-1)  # step out
        _LOG.debug('giving up on finding the surface')
        _LOG.warn(e)
        raise e

    def stepper_approach(self, target_deflection):
        _LOG.info('approach with stepper until deflection > {}'.format(
                target_deflection))
        cd = self.piezo.read_deflection()  # cd = current deflection in bits
        _LOG.debug('single stepping approach')
        while cd < target_deflection:
            _LOG.debug('deflection {} < setpoint {}.  step closer'.format(
                    cd, target_deflection))
            self.stepper.single_step(1)  # step in
            cd = self.piezo.read_deflection()
